@@ -148,22 +148,41 @@ static void _notify_timeout_destroy(NotifyTimeout* nt)
 	g_free(nt);
 }
 
-static void create_stack_for_monitor(NotifyDaemon* daemon, GdkScreen* screen, int monitor_num)
+static int
+_gtk_get_monitor_num (GdkMonitor *monitor)
+{
+	GdkDisplay *display;
+	int n_monitors, i;
+
+	display = gdk_monitor_get_display(monitor);
+	n_monitors = gdk_display_get_n_monitors(display);
+
+	for(i = 0; i < n_monitors; i++)
+	{
+		if (gdk_display_get_monitor(display, i) == monitor) return i;
+	}
+
+	return -1;
+}
+
+static void create_stack_for_monitor(NotifyDaemon* daemon, GdkScreen* screen, GdkMonitor *monitor_num)
 {
 	NotifyScreen* nscreen = daemon->priv->screen;
 
-	nscreen->stacks[monitor_num] = notify_stack_new(daemon, screen, monitor_num, daemon->priv->stack_location);
+	nscreen->stacks[_gtk_get_monitor_num(monitor_num)] = notify_stack_new(daemon, screen, monitor_num, daemon->priv->stack_location);
 }
 
 static void on_screen_monitors_changed(GdkScreen* screen, NotifyDaemon* daemon)
 {
+	GdkDisplay     *display;
 	NotifyScreen* nscreen;
 	int n_monitors;
 	int i;
 
 	nscreen = daemon->priv->screen;
+	display = gdk_screen_get_display (screen);
 
-	n_monitors = gdk_screen_get_n_monitors(screen);
+	n_monitors = gdk_display_get_n_monitors(display);
 
 	if (n_monitors > nscreen->n_stacks)
 	{
@@ -173,7 +192,7 @@ static void on_screen_monitors_changed(GdkScreen* screen, NotifyDaemon* daemon)
 		/* add more stacks */
 		for (i = nscreen->n_stacks; i < n_monitors; i++)
 		{
-			create_stack_for_monitor(daemon, screen, i);
+			create_stack_for_monitor(daemon, screen, gdk_display_get_monitor (display, i));
 		}
 
 		nscreen->n_stacks = n_monitors;
@@ -212,18 +231,20 @@ static void on_screen_monitors_changed(GdkScreen* screen, NotifyDaemon* daemon)
 
 static void create_stacks_for_screen(NotifyDaemon* daemon, GdkScreen *screen)
 {
+	GdkDisplay     *display;
 	NotifyScreen* nscreen;
 	int i;
 
 	nscreen = daemon->priv->screen;
+	display = gdk_screen_get_display (screen);
 
-	nscreen->n_stacks = gdk_screen_get_n_monitors(screen);
+	nscreen->n_stacks = gdk_display_get_n_monitors(display);
 
 	nscreen->stacks = g_renew(NotifyStack*, nscreen->stacks, nscreen->n_stacks);
 
 	for (i = 0; i < nscreen->n_stacks; i++)
 	{
-		create_stack_for_monitor(daemon, screen, i);
+		create_stack_for_monitor(daemon, screen, gdk_display_get_monitor (display, i));
 	}
 }
 
@@ -701,7 +722,7 @@ static void _calculate_timeout(NotifyDaemon* daemon, NotifyTimeout* nt, int time
 
 		theme_set_notification_timeout(nt->nw, timeout);
 
-		glong usec = timeout * 1000;  /* convert from msec to usec */
+		glong usec = timeout * 1000L;  /* convert from msec to usec */
 
 		/*
 		 * If it's less than 0, wrap around back to MAXLONG.
@@ -912,9 +933,12 @@ static GdkPixbuf* _notify_daemon_pixbuf_from_path(const char* path)
 		if (!strncmp (path, "file://", 7))
 		{
 			path += 7;
+
+			/* Unescape URI-encoded, allowed characters */
+			path = g_uri_unescape_string (path, G_URI_RESERVED_CHARS_ALLOWED_IN_PATH);
 		}
 
-			/* Load file */
+		/* Load file */
 		pixbuf = gdk_pixbuf_new_from_file (path, NULL);
 	}
 	else
@@ -1441,13 +1465,9 @@ gboolean notify_daemon_notify_handler(NotifyDaemon* daemon, const char* app_name
 	}
 	else
 	{
-		int monitor_num;
+		GdkMonitor *monitor_id;
 		GdkDisplay *display;
-#if GTK_CHECK_VERSION (3, 20, 0)
 		GdkSeat *seat;
-#else
-		GdkDeviceManager *device_manager;
-#endif
 		GdkDevice *pointer;
 		GdkScreen* screen;
 		gint x, y;
@@ -1460,31 +1480,27 @@ gboolean notify_daemon_notify_handler(NotifyDaemon* daemon, const char* app_name
 		if (g_settings_get_boolean(daemon->gsettings, GSETTINGS_KEY_USE_ACTIVE))
 		{
 			display = gdk_display_get_default ();
-#if GTK_CHECK_VERSION (3, 20, 0)
 			seat = gdk_display_get_default_seat (display);
 			pointer = gdk_seat_get_pointer (seat);
-#else
-			device_manager = gdk_display_get_device_manager (display);
-			pointer = gdk_device_manager_get_client_pointer (device_manager);
-#endif
 
 			gdk_device_get_position (pointer, &screen, &x, &y);
-			monitor_num = gdk_screen_get_monitor_at_point (screen, x, y);
+			monitor_id = gdk_display_get_monitor_at_point (gdk_screen_get_display (screen), x, y);
 		}
 		else
 		{
 			screen = gdk_display_get_default_screen(gdk_display_get_default());
-			monitor_num = g_settings_get_int(daemon->gsettings, GSETTINGS_KEY_MONITOR_NUMBER);
+			monitor_id = gdk_display_get_monitor (gdk_display_get_default(),
+							      g_settings_get_int(daemon->gsettings, GSETTINGS_KEY_MONITOR_NUMBER));
 		}
 
-		if (monitor_num >= priv->screen->n_stacks)
+		if (_gtk_get_monitor_num (monitor_id) >= priv->screen->n_stacks)
 		{
 			/* screw it - dump it on the last one we'll get
 			 a monitors-changed signal soon enough*/
-			monitor_num = priv->screen->n_stacks - 1;
+			monitor_id = gdk_display_get_monitor (gdk_display_get_default(), priv->screen->n_stacks - 1);
 		}
 
-		notify_stack_add_window (priv->screen->stacks[monitor_num], nw, new_notification);
+		notify_stack_add_window (priv->screen->stacks[_gtk_get_monitor_num (monitor_id)], nw, new_notification);
 	}
 
 	if (id == 0)
